/* fs_probe.cpp
Probes the OS and filing system for various characteristics
(C) 2015 Niall Douglas http://www.nedprod.com/
File Created: Nov 2015


Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

#define _CRT_SECURE_NO_WARNINGS 1

#include "include/handle.hpp"
#include "include/storage_profile.hpp"

#include <fstream>
#include <iomanip>

using namespace BOOST_AFIO_V2_NAMESPACE;
#ifdef __linux__
# define file_handle BOOST_AFIO_V2_NAMESPACE::file_handle
#endif

constexpr unsigned permute_flags_max = 4;
static const std::regex sp_preamble{ "(system|storage).*" };

static storage_profile::storage_profile profile[permute_flags_max];

#define RETCHECK(expr) { auto ret=(expr); if(ret.has_error()) { std::cerr << "WARNING: Operation " #expr " failed due to '" << ret.get_error().message() << "'" << std::endl; abort(); } }

int main(int argc, char *argv[])
{
  std::regex torun(".*");
  bool regexvalid = false;
  unsigned torunflags = permute_flags_max-1;
  if (argc > 1)
  {
    try
    {
      torun.assign(argv[1]);
      regexvalid = true;
    }
    catch (...) {}
    if (argc > 2)
      torunflags = atoi(argv[2]);
    if (!regexvalid)
    {
      std::cerr << "Usage: " << argv[0] << " <regex for tests to run> [<flags>]" << std::endl;
      return 1;
    }
  }

  // Force extent allocation before test begins
  {
    io_service service;
    // Create file with O_SYNC
    auto _testfile(file_handle::file(service, "test", handle::mode::write, handle::creation::if_needed, handle::caching::reads));
    if (!_testfile)
    {
      std::cerr << "WARNING: Failed to create test file due to '" << _testfile.get_error().message() << "', failing" << std::endl;
      return 1;
    }
    file_handle testfile(std::move(_testfile.get()));
    std::vector<char> buffer(1024 * 1024);
    RETCHECK(testfile.truncate(buffer.size()));
    file_handle::io_request<file_handle::const_buffers_type> reqs({ std::make_pair(buffer.data(), buffer.size()) }, 0);
    RETCHECK(testfile.write(reqs));
  }
  // File closes, as it was opened with O_SYNC it forces extent allocation
  // Pause as Windows still takes a while
  stl11::this_thread::sleep_for(stl11::chrono::seconds(3));
  std::ofstream results("fs_probe_results.yaml", std::ios::app);
  {
    auto put_time=[](const std::tm* tmb, const char * fmt){
      std::string buffer(256, 0);
      buffer.resize(std::strftime((char *) buffer.data(), buffer.size(), fmt, tmb));
      return buffer;
    };
    std::time_t t = std::time(nullptr);
    results << "---\ntimestamp: " << put_time(std::gmtime(&t), "%F %T %z") << "\n";
  }
  bool first = true;
  for (unsigned flags = 0; flags <= torunflags; flags++)
  {
    if (!flags || !!(flags & torunflags))
    {
      io_service service;
      handle::caching strategy=handle::caching::all;
      switch (flags)
      {
      case 1:
        strategy = handle::caching::only_metadata;  // O_DIRECT
        break;
      case 2:
        strategy = handle::caching::reads;          // O_SYNC
        break;
      case 3:
        strategy = handle::caching::none;           // O_DIRECT|O_SYNC
        break;
      }
      std::cout << "\ndirect=" << !!(flags & 1) << " sync=" << !!(flags & 2) << ":\n";
      auto _testfile(file_handle::file(service, "test", handle::mode::write, handle::creation::open_existing, strategy));
      if (!_testfile)
      {
        std::cerr << "WARNING: Failed to create test file due to '" << _testfile.get_error().message() << "', skipping" << std::endl;
        continue;
      }
      file_handle testfile(std::move(_testfile.get()));
      for (auto &test : profile[flags])
      {
        if (std::regex_match(test.name, torun))
        {
          std::cout << "Running test " << test.name << " ..." << std::endl;
          auto result = test(profile[flags], testfile);
          if (result)
          {
            test.invoke([](auto &i) {
              std::cout << "   " << i.name << " = " << i.value << std::endl;
            });
          }
          else
            std::cerr << "   ERROR running test '" << test.name << "': " << result.get_error().message() << std::endl;
        }
      }
      // Write out results for this combination of flags
      if (first)
      {
        profile[flags].write(results, sp_preamble);
        first = false;
      }
      results << "direct=" << !!(flags & 1) << " sync=" << !!(flags & 2) << ":\n";
      profile[flags].write(results, sp_preamble, 4, true);
      results.flush();
    }
  }

  return 0;
}
