/* utils.hpp
Misc utilities
(C) 2016 Niall Douglas http://www.nedprod.com/
File Created: Jan 2015


Boost Software License - Version 1.0 - August 17th, 2003

Permission is hereby granted, free of charge, to any person or organization
obtaining a copy of the software and accompanying documentation covered by
this license (the "Software") to use, reproduce, display, distribute,
execute, and transmit the Software, and to prepare derivative works of the
Software, and to permit third-parties to whom the Software is furnished to
do so, all subject to the following:

The copyright notices in the Software and this entire statement, including
the above license grant, this restriction and the following disclaimer,
must be included in all copies of the Software, in whole or in part, and
all derivative works of the Software, unless such copies or derivative
works are solely in the form of machine-executable object code generated by
a source language processor.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
*/

#include "../../../utils.hpp"

#include <sys/mman.h>

BOOST_AFIO_V2_NAMESPACE_BEGIN

namespace utils
{
  std::vector<size_t> page_sizes(bool only_actually_available) noexcept
  {
    static spinlock<bool> lock;
    static std::vector<size_t> pagesizes, pagesizes_available;
    stl11::lock_guard<decltype(lock)> g(lock);
    if (pagesizes.empty())
    {
#if defined(__FreeBSD__)
      pagesizes.resize(32);
      int out;
      if(-1==(out=getpagesizes(pagesizes.data(), 32)))
      {
        pagesizes.clear();
        pagesizes.push_back(getpagesize());
        pagesizes_available.push_back(getpagesize());
      }
      else
      {
        pagesizes.resize(out);
        pagesizes_available=pagesizes;
      }
#elif defined(__APPLE__)
      // I can't find how to determine what the super page size is on OS X programmatically
      // It appears to be hard coded into mach/vm_statistics.h which says that it's always 2Mb
      // Therefore, we hard code 2Mb
      pagesizes.push_back(getpagesize());
      pagesizes_available.push_back(getpagesize());
      pagesizes.push_back(2*1024*1024);
      pagesizes_available.push_back(2*1024*1024);      
#elif defined(__linux__)
      pagesizes.push_back(getpagesize());
      pagesizes_available.push_back(getpagesize());
      int ih=::open("/proc/meminfo", O_RDONLY);
      if(-1!=ih)
      {
        char buffer[4096], *hugepagesize, *hugepages;
        buffer[::read(ih, buffer, sizeof(buffer)-1)]=0;
        ::close(ih);
        hugepagesize=strstr(buffer, "Hugepagesize:");
        hugepages=strstr(buffer, "HugePages_Total:");
        if(hugepagesize && hugepages)
        {
          unsigned _hugepages=0, _hugepagesize=0;
          while(*++hugepagesize!=' ');
          while(*++hugepages!=' ');
          while(*++hugepagesize==' ');
          while(*++hugepages==' ');
          sscanf(hugepagesize, "%u", &_hugepagesize);
          sscanf(hugepages, "%u", &_hugepages);
          if(_hugepagesize)
          {
            pagesizes.push_back(((size_t)_hugepagesize)*1024);
            if(_hugepages)
              pagesizes_available.push_back(((size_t)_hugepagesize)*1024);
          }
        }
      }
#else
#warning page_sizes() does not know this platform, so assuming getpagesize() is the best available
      pagesizes.push_back(getpagesize());
      pagesizes_available.push_back(getpagesize());
#endif
    }
    return only_actually_available ? pagesizes_available : pagesizes;
  }
#ifdef _MSC_VER
#pragma warning(pop)
#endif

  void random_fill(char *buffer, size_t bytes)
  {
    static spinlock<bool> lock;
    static int randomfd=-1;
    if(-1==randomfd)
    {
      stl11::lock_guard<decltype(lock)> g(lock);
      randomfd=::open("/dev/urandom", O_RDONLY);
    }
    if(-1==randomfd || ::read(randomfd, buffer, bytes)<bytes)
    {
      BOOST_AFIO_LOG_FATAL_EXIT("afio: Kernel crypto function failed");
      std::terminate();
    }
  }

  namespace detail
  {
    large_page_allocation allocate_large_pages(size_t bytes)
    {
      large_page_allocation ret(calculate_large_page_allocation(bytes));
      int flags=MAP_SHARED|MAP_ANON;
      if(ret.page_size_used>65536)
      {
#ifdef MAP_HUGETLB
        flags|=MAP_HUGETLB;
#endif
#ifdef MAP_ALIGNED_SUPER
        flags|=MAP_ALIGNED_SUPER;
#endif
#ifdef VM_FLAGS_SUPERPAGE_SIZE_ANY
        flags|=VM_FLAGS_SUPERPAGE_SIZE_ANY;
#endif
      }
      if(!(ret.p=mmap(nullptr, ret.actual_size, PROT_WRITE, flags, -1, 0)))
      {
        if(ENOMEM==errno)
          if((ret.p=mmap(nullptr, ret.actual_size, PROT_WRITE, MAP_SHARED|MAP_ANON, -1, 0)))
            return ret;
      }
#ifndef NDEBUG
      else if(ret.page_size_used>65536)
        std::cout << "afio: Large page allocation successful" << std::endl;
#endif
      return ret;
    }
    void deallocate_large_pages(void *p, size_t bytes)
    {
      if(munmap(p, bytes)<0)
      {
        BOOST_AFIO_LOG_FATAL_EXIT("afio: Freeing large pages failed");
        std::terminate();
      }
    }
  }
}

BOOST_AFIO_V2_NAMESPACE_END
